#ifndef TRXTARG_H
/** \file
   Various definitions for target materials.
*/
/*

   Copyright  1992-2015 M.Kraemer, GSI, D-64291 Darmstadt

   $Log: trxtarg.h,v $
   Revision 1.6  2015/01/05 13:46:51  kraemer
   rework/streamline, Auger electrons, mixtures

   Revision 1.5  2009/05/29 12:25:02  kraemer
   rework targets

   Revision 1.4  2008/06/20 14:36:15  kraemer
   x98-libs

   Revision 1.3  2002/11/25 16:37:36  kraemer
   TEA (tri-ethyl-amine) added

   Revision 1.2  2002/01/22 17:31:29  kraemer
   strict for Aix, OS-dependent makefile, bugfix in mdose evaluation

   Revision 1.1  2001/01/09 14:54:42  kraemer
   *** empty log message ***



*/
#define TRXTARG_H

#include "utypes.h"
#include "ulists.h"
#include "ufilemem.h"
#include "m.h"
#include "cmderr.h"
#include "trxproj.h"
                                      /* Common status and action flags */
                                      /* masks */
#define TRXTARG_MASKGLOBAL 0xFF000000 /* global (e.g. debug) */
#define TRXTARG_MASKFUNC   0x00FF0000 /* function specific flags */
#define TRXTARG_MASKQUAL   0x0000FFFF /* for additional qualifiers */
                                      /* */
                                      /* global flags */
#define TRXTARG_DEBUG  0x80000000     /* general check flag */
#define TRXTARG_QUIET  0x10000000     /* suppress error messages */
#define TRXTARG_NOMSG  TRXTARG_QUIET

#define TRXTARG_WITHNONE
                                      /** \struct STRXTARGTAGID
                                          \brief Aux. structure to correlate tag string <=> ID.
                                      */ 
struct STRXTARGTAGID { char *pcTag; int iID; };

                                      /** \struct STRXPARENT
                                          \brief  describes parentage 
                                      */
struct STRXPARENT {
   double dFraction;                  /**< parentage coeff. */
   char cAtom[4];                     /**< parent atom name */
   char *pcCfg;                       /**< parent atomic shell name */
};                                          
                                      /* Auger effects by Cathrin Waelzlein */
				                                  /** \struct STRXAUGER
					                                     \brief describes an auger transition
				                                  */
struct STRXAUGER {
   int iSh2;                          /**< 2nd shell number */
   int iSh3;                          /**< 3rd shell number */
   double dEaug;                      /**< Auger Energie Ebnd1-Ebnd2-Ebnd3 */
   double dprob;                      /**< transition probability */
   double dfyield;                    /**< flourescence yield */
};

                                      /** \struct STRXSHELL 
                                          \brief describes an electron shell 
                                      */
struct STRXSHELL {
   unsigned int uiFlag;               /**< holds flag bits */
#define TRXTARGSHELL_ISVALID   0x00000001  /* flag that shell has been validated */ 
   int in, il;                        /**< quantum numbers */ 
   int    ne;                         /**< # of electrons */
   char *pcCfg;                       /**< shell config */
   size_t ulSpace;                    /**< space occupied */
   double dEbnd,                      /**< experimental binding energy */
          dEkin;                      /**< mean kinetic energy a la Slater */
   double ds;                         /**< shielding constant */
   int nParent;                       /**< # of parent atoms (0 if none) */
   struct STRXPARENT *psParent;       /**< parent atoms */  
   int nAuger, mAuger;                /**< current,max # of Auger transitions if this shell is ionized, 
                                           for calculation with probabilities */
   struct STRXAUGER *psAuger;         /**< [mAuger] Auger transitions for this ionized shell */
};
/*

   Functions to handle a single target shell
   --------------------------------------------

                 
   TRXTargShellDup     ( shto, shfr, , flag)
                       o Duplicate a target shell
                     5 x to test
                       o calls
                         o TRXTargShellFree
                         o TRXTargShellInit
                         o TRXTargShellSpace
                                  
   TRXTargShellFree    ( sh, flag)
                       o Release a target shell's resources
                     4 x to test
                       o calls
                         o TRXTargShellInit
                                  
   TRXTargShellInit    ( sh, flag)
                       o Initialize a target shell
                     3 x to test
                       o calls
                         o TRXTargShellSpace
                 
   TRXTargShellInspect ( sh, flag)
                       o Inspect a target shell
                     1 x to test
                       o calls
                 
   TRXTargShellSpace   ( sh, flag)
                       o Space occupied by a target shell
                     2 x to test
                       o calls
                 
   TRXTargShellValidate( sh, flag)
                       o Verify a target shell
                     6 x to test
                       o calls
                 
*/
                                      /* function prototypes */
struct SCMDERR TRXTargShellDup     ( struct STRXSHELL *, struct STRXSHELL *, unsigned int );
struct SCMDERR TRXTargShellFree    ( struct STRXSHELL *, unsigned int );
struct SCMDERR TRXTargShellInit    ( struct STRXSHELL *, unsigned int );
struct SCMDERR TRXTargShellInspect ( struct STRXSHELL *, unsigned int );
struct SCMDERR TRXTargShellSpace   ( struct STRXSHELL *, unsigned int );
struct SCMDERR TRXTargShellValidate( struct STRXSHELL *, unsigned int );

struct STRXSHELL _TRXTargShellTest ( unsigned int, unsigned int );

enum { TRXTARGCOMP_ATM=0, TRXTARGCOMP_MOL, /* keep this order! */ TRXTARGCOMP_MAX, };
                                      /** \struct STRXTARG 
                                          \brief Describes target material 
                                      */
struct STRXTARG {
   unsigned int uiFlag;               /**< holds flag bits */

                                      /* flag bits: */
#define TRXTARG_ISVALID   0x00000001  /* flag valid content */ 
#define TRXTARG_REFCOPY   0x00000002  /* do not free() */ 
   unsigned int uiCode;               /**< short cut material code number */ 
   unsigned int uiFileVersion;        /**< file version number (on read), yyyymmdd */ 
   int iRefCount;                     /**< how many times this target is referenced */

   size_t ulSpace;                    /**< space occupied by this structure */
   char *pcMaterial;                  /**< material name */
   char *pcComposition;               /**< chemical composition */
   char *pcName;                      /**< stem name */
   struct STRXZA sZA;                 /**< atom species (if no further constituents) */
   double dN;                         /**< relative contribution or number of atoms */
   double dNSum;                      /**< total number of atomic nuclei */
   double dDensity, dMolWeight;       /**< density [g/cm**3], molweight [g] */
   double dnMolecule, dnNuclei;       /**< molecule, nuclei number density */
                                      /* NA / MolWeight */
   int nT;                            /**< # of contributing materials/atoms (from decomposition) */ 
   struct STRXTARG *psT;              /**< [nT] contributing materials/atoms */   
                                      /* molecular shells */  
   int nSh;                           /**< # of shells */
   struct STRXSHELL *psSh;            /**< [nSh] shells */
                                      /* atomic/molecular components (sum formula, linearly ordered) */
   int               nTSum[TRXTARGCOMP_MAX]; 
   struct STRXTARG *psTSum[TRXTARGCOMP_MAX]; 
};
                                      /* commonly used short codes */
enum {
   TRXTARG_NONE=0, 
   TRXTARG_H2O=1, TRXTARG_C, TRXTARG_C2H4, TRXTARG_N2, TRXTARG_O2,
   TRXTARG_PMMA,
   TRXTARG_AL, TRXTARG_NAI, TRXTARG_AGBR, 
   TRXTARG_MYLAR,
   TRXTARG_TEA,
};				      

                                      /* short test whether target material is a mixture */
#define TRXTARGISMIXTURE(t) ( ((t)->nSh <= 0) && ((t)->nT > 0) )
                                      /* short test whether target material is an element */
#define TRXTARGISELEMENT(t) ( ((t)->nSh > 0) && ((t)->nT <= 0) )
                                      /* short test whether target material is a compound */
#define TRXTARGISCOMPOUND(t) ( ((t)->nSh > 0) && ((t)->nT > 0) )

                                      /* target file format version */
/*
#define TRXTARGFILEVERSION 19990513
#define TRXTARGFILEVERSION 20020403
*/
#define TRXTARGFILEVERSION 20131231


/*

   Functions to handle a single target material
   --------------------------------------------

   TRXTargCalc             ( t, projbase, targname, flag)
                           o Try calculate/estimate an atomic target material's properties 
                        16 x to test
                           o to complete
                             . generate cfg, Ebind 
                             ? mixtures, compounds 
                           o calls
                             o TRXTargInit
                             o TRXTargFree
                             o TRXTargShieldSlater
                             o TRXTargEnergies
                             o TRXTargSpace
                             o TRXProj* 
                 
   TRXTargComposition      ( t, projbase, flag)
                           o Decompose target material into constituents 
                        17 x to test
                           o calls:
                             o TRXProj* 
                             o TRXTargInit
                             o TRXTargComposition
                             o TRXTargSpace
                             o TRXTargSelfRef
                       
   TRXTargDup              ( tto, tfr, flag)
                           o Duplicate a target material
                         8 x to test
                           o calls
                             o TRXTargDup
                             o TRXTargFree
                             o TRXTargInit
                             o TRXTargSpace
                             o TRXTargValidate
                             o TRXTargShellDup
                 
   TRXTargEnergySlater     ( t, shell, flag)
                           o Estimate target material's shell energies
                        15 x to test
                           o calls
                             o 

   TRXTargFree             ( t, flag)
                           o Release a target material's resources
                         4 x to test
                           ? consider ref counts
                           o calls
                             o TRXTargInit
                             o TRXTargFree
                             o TRXTargShellFree
                 
   TRXTargImportComposition( t, comp, flag)
                           o Import a target material composition
                         6 x to test
                           o calls
                 
   TRXTargImportMaterial   ( t, material, flag)
                           o Import a target material name
                         5 x to test
                           o calls
                 
   TRXTargInit             ( t, flag)
                           o Initialize a target material
                         3 x to test
                           o calls
                             o TRXTargSpace
                 
   TRXTargInspect          ( t, flag)
                           o Inspect a target material
                         1 x to test
                           o calls
                             o TRXTargInspect
                 
   TRXTargList             ( t, flag)
                           o List a target material
                         9 x to test 
                           o to complete
                             o list all info
                           o calls

   TRXTargMatch            ( t1, t2, flag)
                           o Check if target materials match
                        10 x to test
                           o calls
                 
   TRXTargMolWeight        ( t, flag)
                           o Target material's molweight et al
                        11 x to test
                           o calls
                             o TRXTargMolWeight
                 
   TRXTargRead             ( t, filename, flag)
                           o Read a target material
                        12 x to test
                           o to complete
                             o dedicated converter cfg => n,l
                             ? Validate() here or by caller
                           o calls
                             o TRXTargInit
                             o TRXTargFree
                             o TRXTargImportMaterial
                             o TRXTargImportComposition
                             o TRXTargSpace
                             o TRXTargValidate
                 
   TRXTargSelfRef          ( tmaster, tcheck, flag)
                           o test for self references
                        18 x to test
                           o calls
                 
   TRXTargShieldSlater     ( t, shell, flag)
                           o Estimate target material's shielding constants by Slater's rule
                        14 x to test
                           o calls
                             o 

   TRXTargSpace            ( t, flag)
                           o Space occupied by a target material
                         2 x to test
                           o calls
                             o TRXTargSpace
                 
   TRXTargValidate         ( t, flag)
                           o Verify a target material
                         7 x to test
                           ? various validation levels: after read, after ...
                           o calls
                             o TRXTargShellValidate

   TRXTargWrite            ( t, path, flag)
                           o Write a target material
                        13 x to test
                           o to complete 
                             o dedicated converter n,l => cfg
                             o better (tabulated) layout for cfg
                           o calls


*/
                                      /* function prototypes */
struct SCMDERR TRXTargCalc        ( struct STRXTARG *, struct STRXPROJBASE *,  const char *, unsigned int );
struct SCMDERR TRXTargComposition ( struct STRXTARG *, struct STRXPROJBASE *, unsigned int );

struct SCMDERR TRXTargDup         ( struct STRXTARG *, struct STRXTARG *, unsigned int );
#define TRXTARGDUP_REFCOPY 0x00010000 /* duplicate as a reference copy, i.e. no allocation */
#define TRXTARGDUP_MERGE   0x00020000 /* duplicate via merge, i.e. don't overwrite existing entries */

struct SCMDERR TRXTargEnergySlater( struct STRXTARG *, int, unsigned int );
struct SCMDERR TRXTargFree        ( struct STRXTARG *, unsigned int );
#define TRXTARGFREE_FORCE  0x00010000 /* force free() even if refcounts exist */

struct SCMDERR TRXTargImportComposition( struct STRXTARG *, const char *, unsigned int );
struct SCMDERR TRXTargImportMaterial   ( struct STRXTARG *, const char *, unsigned int );
struct SCMDERR TRXTargInit        ( struct STRXTARG *, unsigned int );

struct SCMDERR TRXTargInspect     ( struct STRXTARG *, unsigned int );
#define TRXTARGINSPECT_CONTRIB 0x00010000 /*  */
#define TRXTARGINSPECT_SHELL   0x00020000 /*  */
#define TRXTARGINSPECT_SUM     0x00040000 /*  */
#define TRXTARGINSPECT_FULL (TRXTARGINSPECT_CONTRIB|TRXTARGINSPECT_SHELL|TRXTARGINSPECT_SUM)

struct SCMDERR TRXTargList        ( struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargMatch       ( struct STRXTARG *, struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargMerge       ( struct STRXTARG *, struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargMolWeight   ( struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargRead        ( struct STRXTARG *, const char *, unsigned int );
struct SCMDERR TRXTargSelfRef     ( struct STRXTARG *, struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargShieldSlater( struct STRXTARG *, int, unsigned int );
struct SCMDERR TRXTargSpace       ( struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargValidate    ( struct STRXTARG *, unsigned int );
struct SCMDERR TRXTargWrite       ( struct STRXTARG *, const char *, unsigned int );

enum { TRXTARGTEST_ATOM=0, TRXTARGTEST_MOLECULE, TRXTARGTEST_MIXTURE, };
struct STRXTARG _TRXTargTest ( unsigned int, unsigned int );

                                      /** \struct STRXTARGBASE 
                                          \brief all defined materials 
                                      */
struct STRXTARGBASE {
   unsigned int uiFlag;               /**< flag bits */
#define TRXTARGBASE_ISVALID   0x00000001  /* flag valid content */ 
   size_t ulSpace;                    /**< space occupied (bytes) */
   int nT, mT;                        /**< current/max # of targets in base */
#define TRXTARGBASE_TARGINC 8         /* mT increment */   
   struct STRXTARG *psT;              /**< [nT/mT] targets */
};

                                      /* target slot increment macro */
#define TRXTARGBASEINCREMENT(psTB,sCE,gError) \
         if ( (psTB)->nT >= (psTB)->mT ) {\
            void *pv = NULL;\
            pv = realloc( (psTB)->psT, (ssize_t) ((psTB)->mT+TRXTARGBASE_TARGINC) * sizeof(*((psTB)->psT)) );\
            if ( pv == NULL ) {\
               (sCE).iRC = ERRTRXPROJ_SYS; goto gError;\
            }\
            (psTB)->psT = pv;\
            memset( (psTB)->psT + (psTB)->mT, 0, TRXTARGBASE_TARGINC * sizeof(*((psTB)->psT)));\
            (psTB)->mT += TRXTARGBASE_TARGINC;\
         }                                

/*
   Functions to handle target base
   -------------------------------

   TRXTargBaseCalc      ( tb, projbase, list, flag)
                        o Try calculate (atomic) target properties 
                     11 x to test
                        o to complete
                          ? refcount/relocations treated correctly?
                        o calls:
                          o TRXTargBaseLocate
                          o TRXTargCalc
                          o TRXTargFree
                          o TRXTargInit
                       
   TRXTargBaseDelete    ( tb, list, flag)
                        o Delete listed targets from base 
                     12 x to test
                        o to complete
                          ? refcount/relocations treated correctly?
                        o calls:
                          o TRXTargFree
                       
   TRXTargBaseFree      ( tb, flag)
                        o Free target base 
                      4 x to test
                        ? consider refcounts
                        o calls:
                          o TRXTargFree
                          o TRXTargBaseInit
                       
   TRXTargBaseInit      ( tb, flag)
                        o Initialize target base
                      3 x to test
                        o calls:
                          o TRXTargBaseSpace
                     
   TRXTargBaseInspect   ( tb, list, flag)
                        o Inspect a target base
                      1 x to test
                        o calls:
                          o TRXTargInspect
                       
   TRXTargBaseList      ( tb, list, flag) 
                        o List targets in base
                      6 x to test
                        o to complete
                          o proper header info
                        o calls:
                          o TRXTargList
                       
   TRXTargBaseLocate    ( tb, target, flag)
                        o Locate a target material in base by structure
                     7b x to test
                        o to complete
                          o locate by composition
                        o calls:
                          o TRXTargMatch
                       
   TRXTargBaseLocateByName( tb, targetname, flag)
                        o Locate a target material in base by name
                     7a x to test
                          o locate by composition
                        o calls:
                     
   TRXTargBaseRead      ( tb, projbase, list, flag)
                        o Read target material files into target material base 
                      9 x to test
                        o to complete
                          o compress identical entries in Sum
                        o calls:
                          o TRXTargBaseLocate
                          o TRXTargShellFromBase
                          o TRXTargRead
                          o TRXTargFree
                          o TRXTargInit
                          o TRXTargComposition
                          o TRXTargMolWeight
                          o TRXTargValidate
                       
   TRXTargShellFromBase ( t, tb, flag)
                        o Completes a target material's shell information from target material base 
                      8 . to test
                          o parentage
                        o calls:
                          o TRXTargBaseLocateByName
                          o TRXTargShieldSlater
                          o TRXTargEnergies
                       
   TRXTargBaseSpace     ( tb, flag) 
                        o Space occupied by a target material
                      2 x to test
                        o calls:
                          o TRXTargSpace
                       
   TRXTargBaseValidate  ( tb, flag)
                        o Verifies validity of a target base
                      5 o to test
                        o calls:
                          o TRXTargValidate
                       
   TRXTargBaseWrite     ( tb, path, list, flag) 
                        o Write target materials from target material base 
                     10 x to test
                        o calls:
                          o TRXTargWrite  

   

*/
struct SCMDERR TRXTargBaseCalc   ( struct STRXTARGBASE *, struct STRXPROJBASE *, const char *, unsigned int );
struct SCMDERR TRXTargBaseDelete ( struct STRXTARGBASE *, const char *, unsigned int );
struct SCMDERR TRXTargBaseFree   ( struct STRXTARGBASE *, unsigned int );
struct SCMDERR TRXTargBaseInit   ( struct STRXTARGBASE *, unsigned int );
struct SCMDERR TRXTargBaseInspect( struct STRXTARGBASE *, const char *, unsigned int );
struct SCMDERR TRXTargBaseValidate( struct STRXTARGBASE *, unsigned int );
struct SCMDERR TRXTargBaseList   ( struct STRXTARGBASE *, const char *, unsigned int );

struct STRXTARG *TRXTargBaseLocate      ( struct STRXTARGBASE *, struct STRXTARG *, unsigned int );
struct STRXTARG *TRXTargBaseLocateByName( struct STRXTARGBASE *, const char *, unsigned int );
#define TRXTARGBASELOCATE_BYCOMPOSITION 0x00010000 /* locate by composition, if material lookup failed */

struct SCMDERR TRXTargBaseRead   ( struct STRXTARGBASE *, struct STRXPROJBASE *, const char *, unsigned int );
struct SCMDERR TRXTargBaseSpace  ( struct STRXTARGBASE *, unsigned int );
struct SCMDERR TRXTargBaseWrite  ( struct STRXTARGBASE *, const char *, const char *, unsigned int );
struct SCMDERR TRXTargShellFromBase      ( struct STRXTARG *, struct STRXTARGBASE *, unsigned int );

struct STRXTARGBASE _TRXTargBaseTest ( unsigned int );
 
                                      /* command processors */
int TRXCmdTarg(
                const char *, const char *, const char *, const char *,
                int32_t *, int32_t *, int32_t *, int32_t *, int32_t *,
                int32_t *, int32_t *
              );

int TRXCmdTargTest(
                const char *, int32_t *
              );

int TRXTargError( struct SCMDERR *, unsigned int );
/*
   --- error codes
*/
enum { ERRTRXTARG_SYS=1, 
       ERRTRXTARG_BADADDR,
       ERRTRXTARG_MISMATCHNAME,
       ERRTRXTARG_BADNAME,
       ERRTRXTARG_BADDENSITY,
       ERRTRXTARG_BADMOLWEIGHT,
       ERRTRXTARG_NOSUCHTARGET,
       ERRTRXTARG_NOFILE,
       ERRTRXTARG_CMDPARNOTYET,
       ERRTRXTARG_BADMATERIAL,
       ERRTRXTARG_NOMOLWEIGHT,
       ERRTRXTARG_BADFILETYPE,
       ERRTRXTARG_BADFILEVERS,
       ERRTRXTARG_ITEMNOTYET,
       ERRTRXTARG_UNKNOWNTAG,
       ERRTRXTARG_SYNTAX,
       ERRTRXTARG_BADSHELL,
       ERRTRXTARG_BADFILELIST,
       ERRTRXTARG_BASEEMPTY,
       ERRTRXTARG_NAMEMISMATCH,
       ERRTRXTARG_COMPMISMATCH,  
       ERRTRXTARG_DENSMISMATCH,  
       ERRTRXTARG_MOLWMISMATCH,  
       ERRTRXTARG_WEIGHTMISMATCH,  
       ERRTRXTARG_NMOLMISMATCH,  
       ERRTRXTARG_NNUCMISMATCH,  
       ERRTRXTARG_SHELLMISMATCH,  
       ERRTRXTARG_CONSTITUENTMISMATCH,  
       ERRTRXTARG_BADCOMPOSITION,
       ERRTRXTARG_BADELEMENT,
       ERRTRXTARG_NOSUCHATOM,
       ERRTRXTARG_NOSUCHSHELL,
       ERRTRXTARG_BADNUMBER,
       ERRTRXTARG_BADPARENT,
       ERRTRXTARG_BADATOM,
       ERRTRXTARG_BADCOMPCHAR,
       ERRTRXTARG_BADCOMPNUM,
       ERRTRXTARG_BADMIXTURE,
       ERRTRXTARG_BADMIXTURESUM,
       ERRTRXTARG_BADMIXTUREREL,
       ERRTRXTARG_BADMIXTURESYNTAX,
       ERRTRXTARG_NOSHELLAVAIL,
       ERRTRXTARG_BADTARGLIST,
       ERRTRXTARG_NOPARENT,
       ERRTRXTARG_PARENTNOSHELL,
       ERRTRXTARG_READUNKNOWNTAG,
       ERRTRXTARG_READUNKNOWNTAGDATA,
       ERRTRXTARG_BADEKIN,
       ERRTRXTARG_NOTATOMIC,
   ERRTRXTARG_BADSHELLCFG,
   ERRTRXTARG_BADSHELLEE,
   ERRTRXTARG_BADSHELLNL,
   ERRTRXTARG_BADSHELLNE,
   ERRTRXTARG_HASREFCOUNT,
   ERRTRXTARG_INVALID,
   ERRTRXTARG_PROJBASEINVALID,
   ERRTRXTARG_SELFREF,
   ERRTRXTARG_READNOSHELL,
   ERRTRXTARG_READNOMATERIAL,
   ERRTRXTARG_WRITEOLDFORMAT,
   ERRTRXTARG_PROJDUPFAILED,
};

#endif
