#ifndef TRXXS_H
#define TRXXS_H
/** \file
   Definitions for cross section tables.
*/
/*

   Copyright  1992-2015 M.Kraemer, GSI Darmstadt

   $Log: trxxs.h,v $
   Revision 1.4  2015/01/05 14:56:16  kraemer
   complete rework for mixtures, nuclear

   Revision 1.3  2009/05/29 12:31:42  kraemer
   reworked target tables, projectiles

   Revision 1.2  2002/11/25 16:39:38  kraemer
   locate macros; outside and emfld pseudo-interactions

   Revision 1.1  2001/01/09 14:54:44  kraemer
   *** empty log message ***


*/

#include <limits.h>

#include "cmderr.h"
#include "trxconst.h"
#include "trxmmap.h"
#include "trxproj.h"
#include "trxtarg.h"

                                      /* Barkas formula */
#define BARKAS(Z0,beta2) ( (Z0) == 1 ) ? (Z0) : ( (Z0) * ( ( 1E0 - exp( -125E0 * sqrt(beta2) * pow( (Z0), -2E0/3E0 ) ) ) ) )
#define TRXXSHASWILD(str) \
   ( ( (str) != NULL ) && ( (strchr(str,'*') != NULL) || (strchr(str,'*') != NULL) ) )
                                      /* Common status and action flags */
                                      /* */
                                      /* masks */
#define TRXXS_MASKGLOBAL 0xFF000000   /* global, bits automatically passed through all function levels */
#define TRXXS_DEBUG      0x80000000   /* general check flag */
#define TRXXS_QUIET      0x40000000   /* suppress error messages */
#define TRXXS_NOMSG      TRXXS_QUIET
                                      /* global function action flags */
#define TRXXS_SHOWDIM    0x01000000   /* for Inspect(), List(), ... : show dim substructure content */
#define TRXXS_SHOWSHELLS 0x02000000   /* for Inspect(), List(), ... : show shell/fragment substructure content */
#define TRXXS_SHOWTABLES 0x04000000   /* for Inspect(), List(), ... : show tables */
#define TRXXS_SHOWALL ( TRXXS_SHOWTABLES | TRXXS_SHOWSHELLS | TRXXS_SHOWDIM )

#define TRXXS_MASKFUNC   0x00FF0000   /* function specific flags */
#define TRXXS_MASKQUAL   0x0000FFFF   /* for additional qualifiers */


                                      /* actions to be handled by common functions */
enum {
TRXXSACTION_NONE=0, 
TRXXSACTION_DELETE=1, TRXXSACTION_DISABLE,  TRXXSACTION_ENABLE, TRXXSACTION_EXPORTGD,
TRXXSACTION_INSPECT, TRXXSACTION_LIST, TRXXSACTION_WRITE, 
TRXXSACTION_MAX
};

                                      /** \struct STRXXSTAG
                                          \brief Aux. structure to correlate tag string <=> ID.
                                      */ 
struct STRXXSTAGID { char *pcTag; int iTag; };

                                      /** \struct STRXXSCALC
                                          \brief Aux. structure to specify cross section calculation input.
                                      */ 
struct STRXXSCALC {
   double dRelChange;                 /**< minimum rel. change between xs table points */
   double dvELim[2];                  /**< range of primary's energies */
   double dvWLim[2];                  /**< range of secondary"s energies */
   double dvALim[2];                  /**< range of secondary's angles (deg) */
   double *pdE, *pdW, *pdA;           /**< list of table breakpoints (optional) */
   char *pcFP, *pcFT;                 /**< list of projectile/target fragments to consider (optional) */
   int nE, nW, nA;                    /**< # of table breakpoints */
   int iMethod;                       /**< cross section calculation method, ID */
   unsigned int uiFlag;               /**< flag bits */ 
#define TRXXSCALC_WABSOLUTE  0x00000001 /* for ions: W's are absolute, not relative to WNorm */ 
#define TRXXSCALC_SKIPUNSUPP 0x00000002 /* skip unsupported cross section */
};    

                                      /* calc. methods */
enum { TRXXS_METHNONE=0, TRXXS_METHDEFAULT=TRXXS_METHNONE,
       TRXXS_METHEMP,
       TRXXS_METHBEA, TRXXS_METHRUDD, 
       TRXXS_METHHKS, TRXXS_METHCDWEIS, 
       TRXXS_METHREL,
       TRXXS_METHRBEB,
       TRXXS_METHBEB,
       TRXXS_METHRFORD,
       TRXXS_METHGSKI,
                                      /* nuclear reactions */
       TRXXS_METHSIHVER,
       TRXXS_METHTRIPATHI,
       TRXXS_METHFSCHMITZ,
};

/*
   TRXXSCalcFree         ( xscalc, flag )
                         o release calculation specs
                       2 x to test
   
   TRXXSCalcInspect      ( xscalc, flag )
                         o inspect calculation specs
                       1 x to test
   
   TRXXSCalcLimitsFromStr( xscalc, str, flag )
                         o determine calculation limits
                       3 x to test
   
   TRXXSCalcMethodFromStr( xscalc, str, flag )
                         o determine calculation method
                       4 o to test
   
*/
struct SCMDERR TRXXSCalcFree         ( struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcInspect      ( struct STRXXSCALC *, unsigned int uiFlag );
                                      /* function specific action flags */
#define TRXXSCALCLIMITSFROMSTR_E    0x00010000  
#define TRXXSCALCLIMITSFROMSTR_W    0x00020000  
#define TRXXSCALCLIMITSFROMSTR_A    0x00040000  
#define TRXXSCALCLIMITSFROMSTR_WABS 0x00080000  
struct SCMDERR TRXXSCalcLimitsFromStr( struct STRXXSCALC *, const char *, unsigned int );
struct SCMDERR TRXXSCalcMethodFromStr( struct STRXXSCALC *, const char *, unsigned int );

/* 
   --- Cross sections for one primary energy 
*/
                                      /** \struct STRXXSE
                                          \brief Cross sections for a single primary energy.
                                      */ 
struct STRXXSE {
   double dE;                         /**< primary Energy,
                                           ions: MeV/u
                                           else: keV
                                      */
   double dEScaled;                   /**< scaled for interpolation */
   double dSigTot;                    /**< total cross section [cm**2] */
   double dSigTotScaled;              /**< scaled for interpolation */
   double dkm;                        /**< mean potential energy transfer */
   double dbeta2;                     /**< (v/c)**2 */
   double dgamma;                     /**< Lorentz factor */
   double dEkin;                      /**< mean kinetic energy of ejected secondaries [keV] */
   double dLET;                       /**< [] */
   double dWNorm;                     /* scale factor for secondary energies
                                         ions: binary peak 
                                      */
   double dvW[4];
   
   int  n1;                           /**< # of secondary energy/angle values */          
   double  *pd1;                      /**< [n1] secondary energy/angle values */
   double  *pdSig1;                   /**< [n1] SDCS */
   double  *pdCum1;                   /**< [n1] SDCS cumulated, normalized */
   int     *pn2;                      /**< [n1] # of secondary angle values */
   double **pd2;                      /**< [n1][n2[]] secondary angle values (cos, 1..-1)*/
   double **pdSig2;                   /**< [n1][n2[]] DDCS */
   double **pdCum2;                   /**< [n1][n2[]] DDCS cumulated, normalized */
#undef NEWIPOL
#if defined(NEWIPOL)
   int n1IPol;                        /**< # of secondary energy/angle values */          
   double  *pd1IPol;                  /**< [n1] secondary energy/angle values */
   double  *pvdCum1[2];               /**< [n1] SDCS cumulated, normalized */
#endif
   struct SCMDERR sCE;                /**< to keep error msg during read */ 
};
                                      /** \struct STRXXSSHELL
                                          \brief Cross sections for a single shell.
                                      */ 
#undef TRXXSSHELL_EQUIVW_AS_IN_CW    /* fall back to version CW used, which treated it incorrectly (my fault, not her's) */
struct STRXXSSHELL {
   unsigned int uiFlag;               /**< flag bits */   
#define TRXXSSHELL_EQUIVW 0x00000001    /* normalized W-tables are equivalent for all primary energies */
   int iSh;                           /**< actual shell number (0=all shells) */
   int nE;                            /**< # of primary energies */
   struct STRXXSE *psE;               /**< [nE] table of cross sections for primary energies */
   struct SCMDERR sCE;                /**< to keep error msg during read */ 
};

                                      /** \struct STRXXSFRAGM
                                          \brief Cross section for nuclear fragment.
                                      */ 
struct STRXXSFRAGM {
   unsigned int uiFlag;               /**< flag bits */   
#define TRXXSFRAGM_ISVALID 0x00000001 /* marks valid cross section */  
   int nE;                            /**< # of primary energies */
   struct STRXXSE *psE;               /**< [nE] table of cross sections for primary energies */
   struct STRXPROJ sP;                /**< fragment properties */ 
   struct SCMDERR sCE;                /**< to keep error msg during read */ 
};
                                      /* common macro to find primary energy bin */

#define TRXXSLOCATEE(iE,psXSELo,psXSEHi,iLo,iHi,psSh,dEScaled)\
   iLo = 0; iHi = (psSh)->nE-1; iE = iLo;\
   for(;;) {\
      if ( ( iE = ( ((iLo)+(iHi)) >> 1 ) ) == iLo ) break;\
      if ( (dEScaled) == (psSh)->psE[iE].dEScaled ) break;\
      if ( (dEScaled) <  (psSh)->psE[iE].dEScaled ) { iHi = iE; }\
      else                                          { iLo = iE; }\
   }\
   psXSELo = (psSh)->psE + (iE);\
   (iE)++; iE = min( (iE), ((psSh)->nE-1) );\
   psXSEHi = (psSh)->psE + (iE);
                                      /* common macro to find SDCS quantity */
#if defined(NEWIPOL)

#define TRXXSLOCATE1(i1,d1,psELo,psEHi,iLo,iHi,ddLo,ddHi,ddEScaled,dq,dd)\
   iLo = 0; iHi = (psELo)->n1IPol-1; i1 = iLo;\
   for(;;) {\
      if ( ( i1 = ( ((iLo)+(iHi)) >> 1 ) ) == iLo ) break;\
      dd = ( (psEHi) -> pvdCum1[0][i1] - (psELo) -> pvdCum1[1][i1] ) * (ddEScaled) +\
                                         (psELo) -> pvdCum1[1][i1];\
      if ( (dq) == (dd) ) break;\
      if ( (dq)  < (dd) ) { iHi = i1; }\
      else                { iLo = i1; }\
   }\
   ddHi = ( (psEHi) -> pvdCum1[0][(i1)+1] - (psELo) -> pvdCum1[1][(i1)+1] ) * (ddEScaled) +\
                                            (psELo) -> pvdCum1[1][(i1)+1];\
   ddLo = ( (psEHi) -> pvdCum1[0][ i1   ] - (psELo) -> pvdCum1[1][ i1   ] ) * (ddEScaled) +\
                                            (psELo) -> pvdCum1[1][ i1   ];\
   d1   = ( (psELo) -> pd1IPol[i1+1] - (psELo) -> pd1IPol[i1  ] )\
        / ( (ddHi) - (ddLo) )\
        * ( (dq  ) - (ddLo) )\
        +   (psELo) -> pd1IPol[i1  ];                                  

#else

#define TRXXSLOCATE1(i1,d1,psELo,psEHi,iLo,iHi,ddLo,ddHi,ddEScaled,dq,dd)\
   iLo = 0; iHi = (psELo)->n1-1; i1 = iLo;\
   for(;;) {\
      if ( ( i1 = ( ((iLo)+(iHi)) >> 1 ) ) == iLo ) break;\
      dd = ( (psEHi) -> pdCum1[i1] - (psELo) -> pdCum1[i1] ) * (ddEScaled) +\
                                     (psELo) -> pdCum1[i1];\
      if ( (dq) == (dd) ) break;\
      if ( (dq)  < (dd) ) { iHi = i1; }\
      else                { iLo = i1; }\
   }\
   ddHi = ( (psEHi) -> pdCum1[(i1)+1] - (psELo) -> pdCum1[(i1)+1] ) * (ddEScaled) +\
                                        (psELo) -> pdCum1[(i1)+1];\
   ddLo = ( (psEHi) -> pdCum1[ i1   ] - (psELo) -> pdCum1[ i1   ] ) * (ddEScaled) +\
                                        (psELo) -> pdCum1[ i1   ];\
   d1   = (psELo) -> pd1[i1  ];\
   if ( ( (ddHi) - (ddLo) ) != 0.0 )\
      d1  += ( (psELo) -> pd1[i1+1] - (psELo) -> pd1[i1  ] )\
           / ( (ddHi) - (ddLo) )\
           * ( (dq  ) - (ddLo) )\
           ;                                  

#endif                                  

/*
   --- data base for one projectile/target/interaction combination
*/
                                      /* cross section dimension */ 
enum {
   TRXXS_NOCS=-1, /* no cross section */
   TRXXS_TCS=0, TRXXS_SDCS=1, TRXXS_DDCS=2,
   TRXXS_MAXDIM
};

/*
   --- 
*/
                                      /** \struct STRXXSCOMPONENT
                                          \brief Cross section composition.
                                      */ 
struct STRXXSCOMPONENT {
   double dWeight;                    /**< weight factor */
   char *pcMaterial;                  /**< to locate "real" cross section by name */
   struct STRXXS *psXS;               /**< cache pointer to "real" cross section, volatile! */
};

/*
   --- 
*/
                                      /** \struct STRXXSDIM
                                          \brief Dimensional cross section.
                                          Normally a DDCS should keep all info.
                                          Additionally separate SDCS and/or TCS are supported.
                                      */ 
struct STRXXSDIM {
   unsigned int uiFlag;               /**< flag bits */   
#define TRXXSDIM_REFCOPY 0x00000004   /* reference copy, do not free() */
#define TRXXSDIM_ISVALID 0x00000001   /* marks valid cross section */  
#define TRXXSDIM_ENABLED 0x00000002   /* marks enabled cross section */  
   int iRefCount;                     /**< # of times this XS is referenced by a cached pointer, don't free() if > 0 */
   size_t ulSpace;                    /**< space counter */
   char *pcInfile;                    /**< cross section input filename */
   int  nSh;                          /**< # of shells to be considered,
                                           may be one more than natural shell numbers,
                                           to hold sum of all shells.
                                      */
   struct STRXXSSHELL *psSh;          /**< [nSh] shell specific cross sections
                                           [last=nSh-1] holds sum of all
                                      */
   int  nF;                           /**< for nuclear interactions: 
                                           # of fragments to be considered,
                                           includes projectile itself [last=nF-1] .
                                      */
   struct STRXXSFRAGM *psF;           /**< nuclear fragments */
   struct STRXMMAP sM;                /**< for memory mapping */
};
/*

   TRXXSDimDup     ( xsdimto, xsdimfr, flag )
                   o Duplicate dimensional cross sectiontables (real copy w/ allocation)
                 5 x to test
                   ? how to support memory mapping
                   ? copy 
                     o infile (no because not read from file)
                     o refcount (no because true copy)
                   ? refcopy
                   o calls:
                     o TRXXSDimFree
                     o TRXXSDimSpace
                     
   TRXXSDimFree    ( xsdim, flag )
                   o release dimensional cross section tables 
                 4 x to test
                   o calls
                     o TRXXSDimInit
                   
   TRXXSDimInit    ( xsdim, flag )
                   o initialize dimensional cross section tables 
                 3 x to test
                   o calls
                     o TRXXSDimSpace
                   
   TRXXSDimIntegrate( xsdim, ia, flag )
                   o initialize dimensional cross section tables 
                 8 o to do
                   o calls
                     o 
                   
   TRXXSDimInspect ( xsdim, flag )
                   o inspect dimensional cross section tables 
                 1 x to test
                   o calls
                   
   TRXXSDimSpace   ( xsdim, flag )
                   o count space occupied by dimensional cross section tables
                 2 o to test
                   o calls
                   
   TRXXSDimShell0  ( xsdim, flag )
                   o Establish shell #0 dimensional cross sections.
                 7 . to test
                     o full test to do
                   o calls
                   
   TRXXSDimValidate( xsdim, flag )
                   o validate dimensional cross section tables
                 6 o to test
                   o calls
                   
*/
struct SCMDERR TRXXSDimDup      ( struct STRXXSDIM *, struct STRXXSDIM *, unsigned int );
struct SCMDERR TRXXSDimFree     ( struct STRXXSDIM *, unsigned int );
struct SCMDERR TRXXSDimInit     ( struct STRXXSDIM *, unsigned int );
struct SCMDERR TRXXSDimIntegrate( struct STRXXSDIM *, int, unsigned int );
struct SCMDERR TRXXSDimInspect  ( struct STRXXSDIM *, unsigned int );
struct SCMDERR TRXXSDimShell0   ( struct STRXXSDIM *, unsigned int );
struct SCMDERR TRXXSDimSpace    ( struct STRXXSDIM *, unsigned int );
struct SCMDERR TRXXSDimValidate ( struct STRXXSDIM *, unsigned int );

struct STRXXSDIM _TRXXSDimTest( unsigned int );             

                                      /** \struct STRXXS
                                          \brief Cross sections for projectile/IA/target specific interaction.
                                      */ 
struct STRXXS {
   unsigned int uiFlag;               /**< flag bits */   
#define TRXXS_REFCOPY    0x00000004   /* reference copy, do not free() */
#define TRXXS_ISVALID    0x00000001   /* marks valid cross section */  
#define TRXXS_ENABLE     0x00000002   /* marks enabled cross section */  
            int iIA;                  /**< interaction type */
   size_t ulSpace;                    /**< space counter */
   int iRefCount;                     /**< # of times this XS is referenced by a cached pointer, don't free() if > 0 */
   int nDim;                          /**< actual cross section dimension (0/1/2 for total,single,double) */
   double dCutOffDiss;                /**< scratch entry for evaluation of "dissipative cutoff"  */

   int nXSC, mXSC;                    /**< current, max # of constituent cross sections (mixtures,nuclear) */
#define TRXXS_INCCOMPONENT 4          /* mXSC increment */   
   struct STRXXSCOMPONENT *psXSC;     /**< [nXSC] references to constituent cross sections,
                                           for mixtures, nuclear */  
   struct STRXXSDIM svDim[TRXXS_MAXDIM]; /**< dimensional cross sections */
   struct STRXTARG sT;                /**< specifies target properties */
   struct STRXPROJ sP;                /**< specifies projectile properties */
};
                                      /* enums */
                                      /* interaction types */     
enum {
                                      /* true interactions (table based) */ 
   TRXXS_NONE=0,
   TRXXS_ELASTIC,                     /*  */    
   TRXXS_IONIZATION, 
   TRXXS_EXCITATION,
   TRXXS_BREMS, 
   TRXXS_ANNIH,
   TRXXS_PHOTO,                       /* 6 */
   TRXXS_COMPTON,
   TRXXS_PAIR,
   TRXXS_NUCLREACT,                   /* nuclear reaction */
   TRXXS_MAXXS,                       /* max # of true interactions */  
                                      /* pseudo interactions */ 
   TRXXS_OUTSIDE=31,
   TRXXS_MAXPSEUDOXS,                 /* max # of true+pseudo interactions:
                                         must be <= 32 !!! (bit positions in 32-bit word)
                                      */  
   TRXXS_ALL,
   TRXXS_EMFLD=TRXXS_OUTSIDE-1,
   TRXXS_BOUNDARY=TRXXS_EMFLD-1,
   TRXXS_CUTOFF=TRXXS_BOUNDARY-1,
   
   TRXXS_BOR=INT_MAX-1, TRXXS_EOR,    /* dummies for begin/end-of-run */
};

/*
#define TRXXS_FILEVERSION 20080825
*/
#define TRXXS_FILEVERSIONWITHNUCLEAR 20130131
#define TRXXS_FILEVERSION TRXXS_FILEVERSIONWITHNUCLEAR
/*
#define TRXXS_CHARS "???lixbafcp"
*/
#define TRXXS_BITOUTSIDE ( 1 << TRXXS_OUTSIDE )
enum {
/*
   TRXXS_BITOUTSIDE    = ( 1L << TRXXS_OUTSIDE ), 
*/
   TRXXS_BITBOUNDARY   = ( 1L << TRXXS_BOUNDARY ), 
   TRXXS_BITCUTOFF     = ( 1L << TRXXS_CUTOFF ), 
   TRXXS_BITELASTIC    = ( 1L << TRXXS_ELASTIC ), 
   TRXXS_BITIONIZATION = ( 1L << TRXXS_IONIZATION ), 
   TRXXS_BITEXCITATION = ( 1L << TRXXS_EXCITATION ), 
   TRXXS_BITBREMS      = ( 1L << TRXXS_BREMS ), 
   TRXXS_BITANNIH      = ( 1L << TRXXS_ANNIH ), 
   TRXXS_BITEMFLD      = ( 1L << TRXXS_EMFLD ), 
};
enum {
   TRXXS_BITPHOTO      = ( 1L << TRXXS_PHOTO ), 
   TRXXS_BITCOMPTON    = ( 1L << TRXXS_COMPTON ), 
   TRXXS_BITPAIR       = ( 1L << TRXXS_PAIR ), 
};
enum {
   TRXXS_BITNUCLREACT  = ( 1L << TRXXS_NUCLREACT ), 
};

/*

   Functions to handle a single cross section
   ------------------------------------------

   TRXXSIATypeFromStr( xs, iactstr, flag )
                     o Convert interaction string to internal id 
                  7b x to test
                     o to complete
                       . allow wildcards
                       
   TRXXSIADimFromStr ( xs, iactstr, flag )
                     o Convert interaction dimension string to internal id 
                  7a x to test
                  
   TRXXSIAFromStr    ( xs, iactstr, flag )
                     o interaction type & dim from given string 
                  7c x to test
                     o to complete
                       . allow wildcards
                     o calls:
                   

   TRXXSAdd        ( xsto, xsfr, flag )
                   o Add a cross section to another cross section
                13 x to test
                   o to complete
                     o noreplace option (for Dim substructure)
                   o calls:
                     o TRXProj*()
                     o TRXTarg*()
                     o TRXXSDimDup
                     o TRXXSFree
                     o TRXXSInit
                     o TRXXSSpace

   TRXXSCalc       ( xs, projbase, proj, ia, targ, xscalc, flag )
                   o steering function to calculate cross sections for selected P/T/IA combination.
                15 x to test
                   o to complete
                     o dimension handling via xscalc?
                     . rc=composedtarget
                       o x,p to do 
                       ? => individual calculation routines
                   o calls:
                     o TRXProj*()
                     o TRXTarg*()
                     o TRXXSInit()
                     o TRXXSFree()
                     o TRXXSIsEquivW()
                     o TRXXSValidate()
                     o TRXXSCalc??
                      
   TRXXSCalcEI     ( xs, proj, targ, xsc, flag )
               15c x to test
                   o to complete
                     o warn in case of default calculation for composed targets
                     o test proper Exportgd for DDCS
                     o H2O
		                     o total xs from subshells
		                     o km from subshells
                     o N2
                     o C2H4
                     o C
                     ? Al
                   o calls:
                                           

   TRXXSCalcEL     ( xs, proj, targ, xsc, flag )
               15a x to test
                   o to complete
                     o warn in case of default calculation for composed targets
                     ? use electron mass etc via psP
                     o allow emin=0
                     o "adapted" breakpoints
                     o Bragg rule for composed targets
                     o partial waves a la Cathrin
                     . E > limit: screened Rutherford/Mott
                     . E < limit: material specific fits
                     ? Escaled as for I,X
                     o GW vs Berger
                   o calls:
                      
   TRXXSCalcEX     ( xs, proj, targ, xsc, flag ) 
               15b x to test
                   o to complete
                     o warn in case of default calculation for composed targets
                     ? use electron mass etc via psP
                     o allow emin=0
                     o "adapted" breakpoints
                     o Bragg rule for composed targets
                     o cleanup
                   o from old cross sections
                     o H2O
		                     o rise < 10eV realistic ?
                     o N2
                       o from old formula
                       o store scaled like fit above
                     o C2H4
                       o from H2O
                       o check w/ W-value and stopping power
                     o C
                       o pickup experimental (?) values
                       o fit 
                     ? Al
                   o relativistic 
                   o consult IAEA
                   o calls:



   TRXXSCalcIN     ( xs, projbase proj, targ, xsc, flag )
                   o nuclear reactions 
               15g o to test
                   o to complete 
                     . common for reaction/fragments ("shell #0")
                     o internal fallback for fragments
                     . methods
                     o cleanup empty cross sections
                     o fragments for projectiles != 4He
                     o "adaptive" break points
                     o format docs
                   o calls


   TRXXSCalcII     ( xs, proj, targ, xsc, flag )
               15d . to test
                   o to complete
                     o warn in case of default calculation for composed targets
                     o possibility to select E, W, A, nE, nW, nA
                     o theories/methods
                       o RBEB complete
                       o BEA
                       o Hansen/Kocbach
                         o empirical 
                         o possibly as user plugin
                       o CDW/EIS  
                       o relativistic
                         o from Moeller
                         o from BEA
                         o from Hansen
                         o Jakubassa
		                     o Wmin = 0
                   o calls:
                      
		 TRXXSCalcIL	    ( xs, xsc, flag)
                   o screened Rutherford/Mott
               15f x to test
                   o to complete
                     o allow external nA
                     o warn in case of default calculation for composed targets
                     o angular binning too fine for output
                       => adapted binning
                      
   TRXXSCalcIX     ( xs, proj, targ, xsc, flag )
               15e x to test
                   o to complete
                     o warn in case of default calculation for composed targets
                   o calls:
                      
                       
   TRXXSCalcXF     ( psXS, psP, psT, psXSC, uiFlag )
                   o to do
                   o dSigma/dOmega of photoelectrons 
		                 o Davisson eq 43 / GW eq 10
		                   (cumulated distrib. valid only for v/c = 1, Z/137 << 1)
                                 ! better angular grid: more dense for angles => 0  
		                 ? < 8 keV: LCAO, Yeh/Lindau At Data NDT 1985 32:1-55
		                 ? own calculation, e.g.: Hulme Proc Roy Soc 133A 381 (1931)
		                 o total cross section
		                   o Davisson Eqs 37 (>2 MeV), 38 ( <2 MeV)
		                     problem for compounds: which Z ???
		                     high energy: only K-Shell => highest Z
                                   ! some E**x scaling  
		                   o Storm/Israel NDT 7 (1970) 565
		                     o 1..200 keV: Brysk/Zerby PR 171 (1968) 292
		                     o > 200 keV : Schmickley/Pratt PR 164 (1967) 104
		                   ? own calculation (Veigele, low energy cross sections)
		                 o Auger electrons:
		                   o H2O: 1 per PE, E=Ek-Elast
		                   o proper fluorescent/Auger yields    
                   o calls:
                      
   TRXXSCalcXC     ( psXS, psP, psT, psXSC, uiFlag ) 
                   o to complete / test
                   o respect pre-allocated tables
                   ? binding energy
                   o evtl more energy steps for high energies
		                 o strange steps in cumulated arrays for very low energies
		                 o E <= 0.01 keV: negative CumA
                       ! CumA < 0, >1,
		                 o form factors for low energies / shell dependencies
                   o scaling to reduce # of table points
                   o calls:
                      
                      
   TRXXSComplete   ( xs, flag )
                   o complete cross section tables 
                 8 . to test
                     o more formal tests to do
                   o to rework
                     . locate / create shell #0
                     . factor 2PI for "ls" cross sections 
                     o integrate cross sections
                     o complete derived quantities in subshells
                     . scaling/norm 
                     ! consistency check (W,A)
                     ? gamma, beta etc for nuclear too
                     . sum of shells, if necessary (to test)
                     . complete km (to test)
                     ? EScaled for E=0
                     ? scaling for photons
                   o calls:
                   
   TRXXSCutOffDiss ( xsto, xsfr, flag )
                   o determine cutoff energy for dissipative interactions.
                     needs XSC->XS to be up-to-date (caller's problem)
                16 x to test
                   o calls:
                     o TRXXSInit

   TRXXSDisable    ( xs, flag )
                   o Disable cross section
                6b x to test
                   o calls:

   TRXXSEnable     ( xs, flag )
                   o Enable cross sections
                6a x to test
                   o calls:

   TRXXSExportgd   ( xs, prefix, flag )
                   o Export cross sections, gd format
                11 x to test
                   o calls:
		                   o TRXXSExportgd()

   TRXXSFree       ( xs, flag )
                   o Release cross section resources
                 4 x to test
                   ? handle refcount
                   o calls:
                     o TRXProj*()
                     o TRXTarg*()
                     o TRXMMap*()
		                   o TRXXSDim*() 
		                   o TRXXSInit() 
                   
   TRXXSInit       ( xs, flag )
                   o Initialize cross section   
                 3 o to test
                   o calls:
                     o TRXProj*()
                     o TRXTarg*()
                     o TRXMMap*()
		                   o TRXXSDimInit() 
		                   o TRXXSSpace() 
                   
   TRXXSInspect    ( xs, flag )
                   o Inspect cross sections 
                 1 x to test
                   o calls:
                     o TRXProjInspect
                     o TRXTargInspect
                     o TRXMMapInspect
                     o TRXXSDimInspect()
                   
   TRXXSIsEquivW   ( xs, eps, flag ) 
                   o tests (and flags) if cross sections have equivalent W-tables
                 9 . to test
                     o more formal tests
                   o calls:

   TRXXSList       ( xs, flag ) 
                   o List some cross section properties (only non-array stuff)                 
                10 x to test
                   o to complete
                     o improve layout
                   o calls:
                     o TRXXSList()
                      
   TRXXSRead       ( xs, filename, psPB, psTB, flag )
                   o read cross section file
                14 x to test 
                     o more extensive tests todo
                   ? memory mapping for binaries
                   ! uFileMemLines() 
                     o empty line problem to be fixed
                     o too slow
                   o calls: 
                     o TRXTarg*()
                     o TRXProj*()
                     o TRXXSIA*FromStr
                     o TRXXSFree()
                     o TRXXSInit()
                     o TRXXSSpace()
                                        
   TRXXSSpace      ( xs, flag ) 
                   o Estimate space for cross sections
                 2 . to test
                     o exact numbers still to check
                   o calls
                     o TRXXSDim*()
                     o TRXTarg*()
                     o TRXProj*()
                    
   TRXXSValidate   ( xs, flag ) 
                   o Validate cross sections 
                 5 x to test
                   o calls:
                     o TRXProjIsValid
                     o TRXTargIsValid
                     o TRXXSDimValidate

   TRXXSWrite      ( xs, prefix, flag )
                   o Write cross sections, native format
                12 x to test
                   o calls:
                     o TRXXSWrite()


*/

                                      /* some function action flags */

#define TRXXSACTION_DDCS 0x00040000 /* action on DDCS only */
#define TRXXSACTION_SDCS 0x00020000 /* action on SDCS only */
#define TRXXSACTION_TCS  0x00010000 /* action on TCS only */

                                      /* function prototypes */

struct SCMDERR TRXXSIAFromStr    ( struct STRXXS *, const char *, unsigned int );
struct SCMDERR TRXXSIATypeFromStr( struct STRXXS *, const char *, unsigned int );
struct SCMDERR TRXXSIADimFromStr ( struct STRXXS *, const char *, unsigned int );
struct SCMDERR TRXXSStrFromIA    ( struct STRXXS *, unsigned int );

#define TRXXSADD_NOREPLACE 0x00020000 /* do not replace existing entries */
#define TRXXSADD_DUPLICATE 0x00100000 /* duplicate (don't clear) input */
struct SCMDERR TRXXSAdd   ( struct STRXXS *, struct STRXXS *, unsigned int );
struct SCMDERR TRXXSCalc( struct STRXXS *, 
                          struct STRXPROJBASE *, struct STRXPROJ *, int , struct STRXTARG *, struct STRXXSCALC *,
                          unsigned int );
struct SCMDERR TRXXSCalcXC( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcXF( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcIN( struct STRXXS *, struct STRXPROJBASE *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcII( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcIL( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcIX( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcEI( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcEL( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSCalcEX( struct STRXXS *, struct STRXPROJ *, struct STRXTARG *, struct STRXXSCALC *, unsigned int );

struct SCMDERR TRXXSComplete( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSDisable ( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSCutOffDiss( struct STRXXS *, struct STRXXS *, unsigned int );
struct SCMDERR TRXXSEnable  ( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSExportgd( struct STRXXS *, const char *, unsigned int );
struct SCMDERR TRXXSFree    ( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSInit    ( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSInspect ( struct STRXXS *, unsigned int );
#if defined(NEWIPOL)
struct SCMDERR TRXXSIntPolSetup ( struct STRXXS *, unsigned int );
#endif
struct SCMDERR TRXXSIsEquivW( struct STRXXS *, double, unsigned int );
struct SCMDERR TRXXSValidate( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSList    ( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSRead    ( struct STRXXS *, const char *, struct STRXPROJBASE *, struct STRXTARGBASE *, unsigned int );
struct SCMDERR TRXXSSpace   ( struct STRXXS *, unsigned int );
struct SCMDERR TRXXSWrite   ( struct STRXXS *, const char *, unsigned int );

struct STRXXS _TRXXSTest( unsigned int );             

/*
   --- for one interaction type
*/
                                      /** \struct STRXXSIACT
                                          \brief Cross sections for a specific interaction.
                                      */ 
struct STRXXSIACT {
   struct STRXPROJ sP;                /**< specifies projectile properties */
   int iIA;                           /**< interaction code number */ 
   unsigned int uiFlag;               /**< flag bits */ 
#define TRXXSIACT_ISVALID 0x00000001  /* marks valid interaction cross section */  
   size_t ulSpace;                    /**< space counter */
   double dCutOffDiss;                /**< scratch entry for evaluation of "dissipative cutoff"  */
   int nXS, mXS;                      /**< current, max # of cross sections */
#define TRXXSIACT_INCXS 8             /* mXS increment */   
   struct STRXXS *psXS;               /**< [mXS] target specific interactions */
};

/*
                       
   TRXXSIActAction     ( xsiact, prefix, targstr, action, flag )
                       o Various interaction cross section actions for selected targets:
                         Disable, Enable, Export in gd format, Inspect, List, Write,
                     7 x to test
                         o Write, Export: incorrect test cases?
                       o to complete
                         o list action
                         . striwld => strwld 
                       o calls:
                         o TRXXSDisable
                         o TRXXSEnable
                         o TRXXSInspect
                         o TRXXSList
                         o TRXXSExportgd
                         o TRXXSWrite
   
   TRXXSIActAdd        ( xsiactto, xsiactfr, flag )
                       o Add an interaction cross section to interaction cross sections
                    10 x to test
                       o to complete
                         . cache XSC => TRXXSIActUpdate
                         o noreplace
                         o add to existing
                       o calls:
                         o TRXProj*()
                         o TRXXSIActXSAdd()
                         o TRXXSIActInit()
                         o TRXXSIActSpace()
                         o TRXXSIActValidate()
                         o TRXXSIActUpdate()
                       
   TRXXSIActCalc       ( xsia, projbase, targbase, proj, iact, targstr, xsc, flag )                    
                       o Steering function to calculate interaction cross sections 
                         for specified projectile/interaction/target combination
                    12 x to test
                       o to complete
                         . dim handling
                         o improve on kludge for isotope handling
                         o iact list
                         o refcounts for composed targets?
                         ! for molecular targets in compositions:
                           need atomic decomposition for default cross sections
                       o calls:
                         o TRXProj*()
                         o TRXXSCalc
                         o TRXXSFree
                         o TRXXSInit()
                         o TRXXSIsValid()
                         o TRXXSIActInit
                         o TRXXSIActXSAdd
                         o TRXXSIActTargLocate
                         o TRXXSIActFree
                         o TRXXSIActSpace()
                         o TRXXSIActValidate()
                         o TRXXSIActUpdate()
                     
   TRXXSIActCutOffDiss ( xsiato, xsiafr, flag )
                       o determine cutoff energy for dissipative interactions.
                    13 x to test
                       o calls:
                         o TRXXSCutOffDiss
                         o TRXXSIActInit

   TRXXSIActDelete     ( xsia, targstr, flag )
                       o Delete interaction cross sections for selected targets
                    11 x to test
                       o to complete
                         ? cross references
                         . striwld() => strwld()
                       o calls:
                         o TRXXSFree()
                         o TRXXSIActSpace()
                         o TRXXSIActUpdate()
   
   TRXXSIActFree       ( xsia, flag )
                       o Release an interaction specific cross section's ressources.
                     4 x to test
                       o to complete
                         o account for cross references
                       o calls:
                         o TRXProj*()
                         o TRXXSFree()
                         o TRXXSIActInit()
                       
   TRXXSIActInit       ( xsia, flag )
                       o Initialize an interaction specific cross section.
                     3 x to test
                       o calls:
                         o TRXProj*()
                         o TRXXSIActSpace()
                       
   TRXXSIActInspect    ( xsia, flag )
                       o Inspect an interaction specific cross section.
                     1 x to test
                       o calls:
                         o TRXProj*()
                         o TRXXSInspect()
                       
   TRXXSIActSpace      ( xsia, flag )
                       o Estimate space occupied by an interaction cross section.
                     2 x to test
                       o calls:
                         o TRXProj*()
                         o TRXXSSpace()
                       
   TRXXSIActTargLocateByName( xsia, targstr, flag )
   TRXXSIActTargLocate ( xsia, targ, flag )
                       o locate target cross section.
                     6 x to test
                       o calls
                         o TRXTargMatch
                     
   TRXXSIActUpdate     ( xsia, flag )
                       o Locate and cache component cross sections.
                     8 . to test
                         o real case to test
                       o calls:

   TRXXSIActValidate   ( xsia, flag )
                       o Verify interaction specific cross section's validity.
                     5 x to test
                       o calls:
                         o TRXProj*()
                         o TRXXSValidate()
                       
   TRXXSIActXSAdd      ( xsia, xs, flag )
                       o Add a single cross section to an interaction cross section.
                     9 x to test
                         . cache XSC
                         o noreplace
                         o add to existing
                       o calls:
                         o TRXProj*()
                         o TRXTarg*()
                         o TRXXSInit
                         o TRXXSAdd
                         o TRXXSIActSpace()
                         o TRXXSIActValidate()
                         o TRXXSIActUpdate()
*/
                                       /* some function action flags */
#define TRXXSIACTACTION_DDCS 0x00040000 /* action on SDCS only */
#define TRXXSIACTACTION_SDCS 0x00020000 /* action on SDCS only */
#define TRXXSIACTACTION_TCS  0x00010000 /* action on TCS only */

                                       /* function prototypes */
struct SCMDERR TRXXSIActAction  ( struct STRXXSIACT *, const char *, const char *, unsigned int, unsigned int);
#define TRXXSIACTADD_NOREPLACE 0x00020000 /* replace existing entries */
#define TRXXSIACTADD_DUPLICATE 0x00100000 /* duplicate (don't clear) input */
struct SCMDERR TRXXSIActAdd     ( struct STRXXSIACT *, struct STRXXSIACT *, unsigned int);
struct SCMDERR TRXXSIActCalc    ( struct STRXXSIACT *, 
                                  struct STRXPROJBASE *, struct STRXTARGBASE *, 
                                  struct STRXPROJ *, int, const char *, struct STRXXSCALC *, 
                                  unsigned int );
struct SCMDERR TRXXSIActCutOffDiss( struct STRXXSIACT *, struct STRXXSIACT *, unsigned int);
struct SCMDERR TRXXSIActDelete  ( struct STRXXSIACT *, const char *, unsigned int);
struct SCMDERR TRXXSIActFree    ( struct STRXXSIACT *, unsigned int);
struct SCMDERR TRXXSIActInit    ( struct STRXXSIACT *, unsigned int);
struct SCMDERR TRXXSIActInspect ( struct STRXXSIACT *, unsigned int);
struct SCMDERR TRXXSIActSpace   ( struct STRXXSIACT *, unsigned int);
struct STRXXS *TRXXSIActTargLocate( struct STRXXSIACT *, struct STRXTARG *, unsigned int );
struct STRXXS *TRXXSIActTargLocateByName( struct STRXXSIACT *, const char *, unsigned int );
struct SCMDERR TRXXSIActUpdate  ( struct STRXXSIACT *, unsigned int);
struct SCMDERR TRXXSIActValidate( struct STRXXSIACT *, unsigned int);

#define TRXXSIACTXSADD_NOREPLACE 0x00020000 /* replace existing entries */
#define TRXXSIACTXSADD_DUPLICATE 0x00100000 /* duplicate (don't clear) input */
struct SCMDERR TRXXSIActXSAdd   ( struct STRXXSIACT *, struct STRXXS *, unsigned int);

struct STRXXSIACT _TRXXSIActTest( unsigned int );             

/*
   --- Cross sections for one projectile
*/
                                      /** \struct STRXXSPROJ
                                          \brief Cross sections for a specific projectile.
                                      */ 
struct STRXXSPROJ {
   struct STRXPROJ sP;                /**< specifies projectile properties */
   unsigned int uiFlag;               /**< flag bits */ 
#define TRXXSPROJ_ISVALID 0x00000001  /* marks valid interaction cross section */  
   size_t ulSpace;                    /**< space counter */
   double dCutOffDiss;                /**< scratch entry for evaluation of "dissipative cutoff"  */
   int nXSIA, mXSIA;                  /**< current, max # of interactions */
#define TRXXSPROJ_INCIACT 4           /* mXSIA increment */   
   struct STRXXSIACT *psXSIA;         /**< [mXSIA] cross sections */
};
                         
/*

   TRXXSProjAction     ( xsp, prefix, iactstr, targstr, action, flag )
                       o Various projectile cross section actions for selected iact/targ combinations:
                         Disable, Enable, Export in gd format, Inspect, List, Write,
                     7 x to test
                       o to complete
                         ? ia="ilx" => would need ia-bits
                         ? TRXXSIActMatch()
                         o list action
		                     o calls
                         o TRXXSInit()
                         o TRXXSIAFromStr
                         o TRXXSIActAction()
		       
   TRXXSProjAdd        ( xspto, xspfr, flag ) 
                       o Add a projectile cross section to projectile cross sections
                    10 x to test
                         o non-duplicate
                       o to complete  
                         o action flags: noreplace
                       o calls:
                         o TRXProjMatch()
                         o TRXXSProjInit
                         o TRXXSProjSpace
                         o TRXXSProjXSIActAdd
                         o TRXXSProjUpdate
                         o TRXXSProjValidate

   TRXXSProjCalc       ( xsp, projbase, targbase, proj, iactstr, targstr, xsc, flag )
                       o Calculate projectile cross sections for selected iact/targ combination.
                    13 x to test
                       o to complete
                         o iact list/wildcard handling
                         o dim handling?
                         o replace/duplicate flags
                       o calls:
                         o TRXXSIATypeFromStr()
                         o TRXXSInit()
                         o TRXXSIActCalc()
                         o TRXXSIActFree()
                         o TRXXSIActInit()
                         o TRXXSProjInit
                         o TRXXSProjFree
                         o TRXXSProjXSIActAdd
                         o TRXXSProjSpace
                         
   TRXXSProjCutOffDiss ( xspto, xspfr, flag )
                       o determine cutoff energy for dissipative interactions.
                    12 x to test
                       o calls:
                         o TRXXSIActCutOffDiss

   TRXXSProjDelete     ( xsp, iactstr, targstr, flag )
                       o Delete projectile cross section for selected iact/targ combination
                    11 x to test
                       o to complete
                         ? use TRXXSIActMatch()
                       o calls:
                         o TRXXSIATypeFromStr
                         o TRXXSIActDelete()
                         o TRXXSIActFree
                         o TRXXSIActIsValid
                         o TRXXSProjFree
                         o TRXXSProjSpace
                         o TRXXSProjUpdate
			   
                                                 
   TRXXSProjFree       ( xsp, flag )
                       o Release projectile cross section resources.
                     4 x to test
                       o calls:
                         o TRXProjFree()
                         o TRXXSIActFree()
                         o TRXXSProjInit
                         
   TRXXSProjInit       ( xsp, flag )
                       o Initialize projectile cross section.
                     3 x to test
                       o calls:
                         o TRXProjInit()
                         o TRXXSProjSpace
                                                   
   TRXXSProjInspect    ( xsp, flag )
                       o Inspect a projectile cross section.
                     1 x to test
                       o calls:
                         o TRXProj*()
                         o TRXXSIAct*()
                       
   TRXXSProjSpace      ( xsp, flag ) 
                       o count space
                     2 x to test
                       o calls:
                         o TRXProjSpace()
                         o TRXXSIActSpace()
                    
   TRXXSProjUpdate     ( xsia, flag )
                       o Locate and cache component cross sections.
                     6 x to test
                         o real case
                       o calls:

   TRXXSProjValidate   ( xsp, flag )
                     5 x to test
                       o calls:
                         o TRXProjValidate()
                         o TRXXSIActValidate()

   TRXXSProjXSAdd      ( xsp, xs, flag )
                       o Adds a single cross section to a projectile cross section
                     8 x to test
                         o non-duplicate
		                     o to complete
                          o action flags: duplicate, noreplace
		                     o calls
                         o TRXProjDup()
                         o TRXProjMatch()
                         o TRXXSInit
                         o TRXXSIActXSAdd()
                         o TRXXSProjSpace
                         o TRXXSProjUpdate
                         o TRXXSProjValidate
		       
   TRXXSProjXSIActAdd  ( xsp, xsia, flag ) 
                       o Add an interaction cross section to projectile cross sections
                     9 x to test
                         o non-duplicate
                       o to complete:
                         o action flags: duplicate, noreplace
                       o calls:
                         o TRXProjDup()
                         o TRXProjMatch()
                         o TRXXSIActAdd()
                         o TRXXSIActInit
                         o TRXXSProjSpace
                         o TRXXSProjUpdate
                         o TRXXSProjValidate
                       
*/
                                       /* function prototypes */

struct SCMDERR TRXXSProjAction   ( struct STRXXSPROJ *, const char *, const char *, const char *, unsigned int, unsigned int );
#define TRXXSPROJADD_NOREPLACE   0x00020000 /* do not replace existing entries */
#define TRXXSPROJADD_DUPLICATE   0x00100000 /* do not erase but duplicate input proj */
struct SCMDERR TRXXSProjAdd      ( struct STRXXSPROJ *, struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjCalc     ( struct STRXXSPROJ *, 
                                   struct STRXPROJBASE *, struct STRXTARGBASE *,
                                   struct STRXPROJ *, const char *, const char *, struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSProjCutOffDiss( struct STRXXSPROJ *, struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjDelete   ( struct STRXXSPROJ *, const char *, const char *, unsigned int );
struct SCMDERR TRXXSProjFree     ( struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjInit     ( struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjInspect  ( struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjSpace    ( struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjUpdate   ( struct STRXXSPROJ *, unsigned int );
struct SCMDERR TRXXSProjValidate ( struct STRXXSPROJ *, unsigned int );

#define TRXXSPROJXSADD_NOREPLACE   0x00020000 /* do not replace existing entries */
#define TRXXSPROJXSADD_DUPLICATE   0x00100000 /* do not erase but duplicate input proj */
struct SCMDERR TRXXSProjXSAdd    ( struct STRXXSPROJ *, struct STRXXS *, unsigned int );

#define TRXXSPROJXSIACTADD_NOREPLACE   0x00020000 /* do not replace existing entries */
#define TRXXSPROJXSIACTADD_DUPLICATE   0x00100000 /* do not erase but duplicate input proj */
struct SCMDERR TRXXSProjXSIActAdd( struct STRXXSPROJ *, struct STRXXSIACT *, unsigned int );

struct STRXXSPROJ _TRXXSProjTest( unsigned int );             



/*
   --- entire cross section data base 
*/
                                      /* */
                                      /** \struct STRXXSBASE
                                          \brief Cross section data base.
                                      */ 
struct STRXXSBASE {
   size_t ulSpace;                    /**< space counter */
   unsigned int uiFlag;               /**< flag bits */
#define TRXXSBASE_ISVALID 0x00000001  /* marks valid interaction cross section */  
   int nXSP, mXSP;                    /**< current/max # of projectiles in base */
#define TRXXSBASE_INCPROJ 4           /* mXSP increment */   
   struct STRXXSPROJ *psXSP;          /**< [mXSP] projectiles */
};
/*
   Functions to handle cross section base
   --------------------------------------
   
   TRXXSBaseAction   ( xsbase, prefix, projstr, iactstr, targstr, action, flag )
                     o Various cross section base actions for selected proj/iact/targ combinations:
                       Disable, Enable, Export in gd format, Inspect, List, Write,
                   6 x to test
                     o to complete
                       o list action
		                   o calls
                       o TRXProj*()
                       o TRXXSProjAction()
		       
   TRXXSBaseCalc     ( xsbase, projbase, targbase, projstr, iactstr, targstr, xscalc, flag )
                     o Calculate cross sections for specified proj/iact/targ
                  13 x to test
                     o to complete
                       ? projectile list with wildcards, against what? 
		                   o calls
                       o TRXProj*()
                       o TRXTarg*()
                       o TRXXSProj*()
                       o TRXXSBaseProjLocate
                       o TRXXSBaseXSProjAdd()

   TRXXSBaseDelete   ( xsbase, projstr, iactstr, targstr, flag)
                     o Delete cross sections for specified proj/iact/targ
                  12 x to test
                     ? pass TRXXSProjDelete return code
		                   o calls
                       o TRXProj*()
                       o TRXXSProj*()
                       o TRXXSBaseFree
                       o TRXXSBaseSpace
                       o TRXXSBaseUpdate
		       
   TRXXSBaseFree     ( xsbase, flag )
                     o Release cross section data base's resources.
                   4 x to test
		                   o calls
		                     o TRXXSProjFree()
                       o TRXXSBaseInit()
		       
   TRXXSBaseInit     ( xsbase, flag )
                     o Initialize a cross section data base.
                   3 x to test
		                   o calls
                       o TRXXSBaseSpace

   TRXXSBaseInspect  ( xsbase, flag )
                     o Inspect a cross section data base.
                   1 x to test
		                   o calls
                       o TRXXSProjInspect

   TRXXSBaseIsProjConsistent( xsbase, proj, flag )
                     o Verify projectile consistency
                     ? to do
		                   o calls
		        
   TRXXSBaseProjLocate( xsbase, proj, flag )
                     o Locate projectile cross section in cross section data base
                   7 x to test
                       o more realistic test to do
                     o calls 
                       o TRXProjMatch
			
   TRXXSBaseRead     ( xsbase, projbase, targbase, filelist, flag )
                     o Read cross sections from specified files
                  11 x to test
                     ? support reading compositions, provided constituents exist
		                   o calls
                       o TRXXSInit()
                       o TRXXSRead()
                       o TRXXSComplete()
                       o TRXXSIsValid
                       o TRXXSIsEquivW
                       o TRXXSFree()
                       o TRXXSBaseXSAdd()

   TRXXSBaseSpace    ( xsbase, flag) 
                     o Calculate space occupied by cross section data base
                   2 x to test 
                     o calls 
                       o TRXXSProjSpace()
                    
   TRXXSBaseUpdate   ( xsia, flag )
                     o Locate and cache references to component cross sections.
                   8 x to test
                       o real case todo
                     o calls:
                       o TRXXSProjUpdate

   TRXXSBaseValidate ( xsbase, flag )
                     o Check validity of cross section data base
                   5 x to test
		                   o calls
                       o TRXXSProjValidate()
		        
   TRXXSBaseXSAdd    ( xsbase, xs, flag )
                     o Adds a single cross section to cross section base
                   9 x to test
                     o to complete
                       o action flags: duplicate, noreplace
		                   o calls
                       o TRXXSInit()
                       o TRXXSIsValid()
                       o TRXXSProjXSAdd()
                       o TRXXSBaseProjLocate()
                       o TRXXSBaseSpace                
                       o TRXXSBaseUpdate
	                      o TRXXSBaseValidate
		       
   TRXXSBaseXSProjAdd( xsbase, xsp, flag )
                     o Adds a projectile cross section to cross section base
                  10 x to test
                     o to complete
                       o action flags: duplicate, noreplace
		                   o calls
                       o TRXXSProjInit
                       o TRXXSProjIsValid
                       o TRXXSProjAdd()
                       o TRXXSBaseSpace
                       o TRXXSBaseProjLocate()
	                      o TRXXSBaseUpdate
	                      o TRXXSBaseValidate
	       
*/
                                      /* function prototypes */
struct SCMDERR TRXXSBaseAction   ( struct STRXXSBASE *, const char *, const char *, const char *, const char *, unsigned int , unsigned int );
struct SCMDERR TRXXSBaseCalc     ( struct STRXXSBASE *, 
                                   struct STRXPROJBASE *, 
                                   struct STRXTARGBASE *,
                                   const char *, const char *, const char *, 
                                   struct STRXXSCALC *, unsigned int );
struct SCMDERR TRXXSBaseDelete   ( struct STRXXSBASE *, const char *, const char *, const char *, unsigned int );
struct SCMDERR TRXXSBaseFree     ( struct STRXXSBASE *, unsigned int );
struct SCMDERR TRXXSBaseInit     ( struct STRXXSBASE *, unsigned int );
struct SCMDERR TRXXSBaseInspect  ( struct STRXXSBASE *, unsigned int );
struct SCMDERR TRXXSBaseRead     ( struct STRXXSBASE *, struct STRXPROJBASE *, struct STRXTARGBASE *, const char *, unsigned int );
struct SCMDERR TRXXSBaseSpace    ( struct STRXXSBASE *, unsigned int );
struct SCMDERR TRXXSBaseUpdate   ( struct STRXXSBASE *, unsigned int );
struct SCMDERR TRXXSBaseValidate ( struct STRXXSBASE *, unsigned int );

#define TRXXSBASEXSADD_NOREPLACE       0x00020000 /* do not replace existing entries */
#define TRXXSBASEXSADD_DUPLICATE       0x00100000 /* do not erase but duplicate input proj */
struct SCMDERR TRXXSBaseXSAdd    ( struct STRXXSBASE *, struct STRXXS *, unsigned int );

#define TRXXSBASEXSPROJADD_NOREPLACE   0x00020000 /* do not replace existing entries */
#define TRXXSBASEXSPROJADD_DUPLICATE   0x00100000 /* do not erase but duplicate input proj */
struct SCMDERR TRXXSBaseXSProjAdd( struct STRXXSBASE *, struct STRXXSPROJ *, unsigned int );

struct STRXXSPROJ *TRXXSBaseProjLocate( struct STRXXSBASE *, struct STRXPROJ *, unsigned int );

struct STRXXSBASE _TRXXSBaseTest( unsigned int );             

                                      /* command processors */
int TRXCmdXS(
              const char *, const char *, const char *,
              const char *, const char *, const char *, 
              int32_t *, int32_t *, int32_t *, int32_t *, int32_t *, int32_t *,
              int32_t *, int32_t *, int32_t *, int32_t *, int32_t *,
              int32_t *, int32_t *, int32_t *,
              const char *, const char *, const char *, const char *, 
              const char *, 
              double *,
              int32_t * 
            );
            
int TRXCmdXSTest      ( const char *, int32_t * );

/*
   --- error handling
*/

int TRXXSError( struct SCMDERR *, unsigned int );

enum { ERRTRXXS_SYS=1, 
       ERRTRXXS_BADADDR,
       ERRTRXXS_BADNUMBER,
       ERRTRXXS_BADEDIM,
       ERRTRXXS_BADPROJ,
       ERRTRXXS_BADTARG,
       ERRTRXXS_BADNSH,
       ERRTRXXS_BADNE,
       ERRTRXXS_BADE,
       ERRTRXXS_BADSIGTOT,
       ERRTRXXS_BADWDIM,
       ERRTRXXS_BADADIM,
       ERRTRXXS_BADSIG2,
       ERRTRXXS_NOSUCHPROJ,
       ERRTRXXS_PROJDUPFAILED,
       ERRTRXXS_TARGDUPFAILED,
       ERRTRXXS_NOSUCHMETHOD,
       ERRTRXXS_BADW,
       ERRTRXXS_BADA,
       ERRTRXXS_BADIACT,
       ERRTRXXS_BADDIMENSION,
       ERRTRXXS_BADXSPROJ,
       ERRTRXXS_BADXSTARG,
       ERRTRXXS_BADITEMLIST,
       ERRTRXXS_NOSUCHMATERIAL,
       ERRTRXXS_EOUTOFRANGE,
       ERRTRXXS_NOTCOMPTON,
       ERRTRXXS_BADCUMW,
       ERRTRXXS_BADCUMA,       
       ERRTRXXS_TARGMATCHFAILED,
       ERRTRXXS_PROJMATCHFAILED,
       ERRTRXXS_NOVALIDXS,
       ERRTRXXS_BASEEMPTY,
       ERRTRXXS_BASENOTVALID,
       ERRTRXXS_BADLIST,
       ERRTRXXS_BADFILELIST,
       ERRTRXXS_RUDDUNSUPPTARG,       
       ERRTRXXS_READUNKNOWNTAG,
       ERRTRXXS_READSYNTAX,
       ERRTRXXS_READNODATA,
       ERRTRXXS_READTOOFEWDATA,
       ERRTRXXS_READBADFILETYPE,
       ERRTRXXS_READBADFILEVERSION,
       ERRTRXXS_READBADNUMBER,
       ERRTRXXS_READBADTARGET,
       ERRTRXXS_READBADPROJECTILE,
       ERRTRXXS_READBADSHELL,
       ERRTRXXS_READBADNSHELL,
       ERRTRXXS_READBADE,
       ERRTRXXS_READBADW,
       ERRTRXXS_READBADXSDATA,
       ERRTRXXS_READDUPSHELLNO,
       ERRTRXXS_READDUPE,
       ERRTRXXS_READDUPW,
       ERRTRXXS_BADSDCS,
       ERRTRXXS_BADSDCSVAL,
       ERRTRXXS_BADSDCSCUM,
       ERRTRXXS_BADDDCS,
       ERRTRXXS_BADDDCSVAL,
       ERRTRXXS_BADDDCSCUM,
       ERRTRXXS_TARGCOMPLETEFAILED,
       ERRTRXXS_BADSHELLNO,
       ERRTRXXS_TARGCALCFAILED,
       ERRTRXXS_BADTARGCONSTITUENTS,
       ERRTRXXS_PROJCOPYFAILED,
       ERRTRXXS_TARGCOPYFAILED,
   ERRTRXXS_TARGISMIXTURE,
   ERRTRXXS_NOSUCHTARGXS,
   ERRTRXXS_GSUNSUPPTARG,
   ERRTRXXS_READBADMOD,
   ERRTRXXS_READBADFRAGM,
   ERRTRXXS_READBADNFRAGM,
   ERRTRXXS_READDUPFRAGM,
   ERRTRXXS_BADNF,
   ERRTRXXS_TARGNOTANUCLIDE,
   ERRTRXXS_TARGMISMATCH,
   ERRTRXXS_BADACTION,
   ERRTRXXS_NOSUCHIACT,
   ERRTRXXS_IACTMATCHFAILED,
   ERRTRXXS_BADXSIACT,
   ERRTRXXS_COMPOSEDATOMIC,
   ERRTRXXS_COMPOSEDMOLECULAR,
   ERRTRXXS_ISREFERENCED,
   ERRTRXXS_XSDIMINVALID,
   ERRTRXXS_XSINVALID,
   ERRTRXXS_XSIACTINVALID,
   ERRTRXXS_XSPROJINVALID,
   ERRTRXXS_PROJBASEINVALID,
   ERRTRXXS_TARGBASEINVALID,
   ERRTRXXS_PROJINVALID,
   ERRTRXXS_TARGINVALID,
   ERRTRXXS_XSCOMPINVALID,
   ERRTRXXS_XSBASEINVALID,
   ERRTRXXS_IACTUNSUPP,
   ERRTRXXS_IOFLO,
};

#endif
